#include <stdlib.h>
#include <stdio.h>
#include <stdio_ext.h>
#include <pthread.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include "exchange.h"
#include "main.h"
#include "astrotime.h"
#include "version.h"

const char *vers = AVRUP_VERSION_TEXT;

unsigned short code[FULL_SIZE];
unsigned short buff[PAGE_SIZE_8+PAGE_SIZE_16];
unsigned char  data[MAX_SIZE];
unsigned char  rdata[MAX_SIZE];

// uC-dependent parameters ATMega8 by default:
int prog_date = PROG_DATE_8;
int page_size = PAGE_SIZE_8;
int line_pars = LINE_PARS_8;

FILE   *datfile;
FILE   *cfgfile;
char   *fname;
char    device[256]="/dev/mcua0";
int new_address = 0;
int new_rate = 0;
int new_ubrr = 0;
int new_serialno = 0;
int mod_address = MODULE;
int fosc = FOSC14;
int ibaude_rate = -1;
int rc_calibrate = -1;
int regime = WRITE;
int high_dump = FULL_SIZE;

int ubrr_to_rate( int rbr) {
    return( rbr < 0 ? fosc/8 : fosc/16/( rbr + 1 ) );
}

int rate_to_ubrr( int rate ) {
    return( rate > 1200 ? -1 : (int)round(fosc/16.0/rate) - 1 );
}

AstroDateTime utc;

int main( int ac, char **av){

    int i;
    unsigned int adr, adr_last, adr_begin = 0xffff;
    unsigned int content, acc, unused = 0;
    unsigned int page, part;

    fprintf( stdout, "\n                AVR UPdate Ver.%s\n", vers );
    opterr = 0;
    while( (i = getopt(ac, av, ":a:b:D:r:N:R:f:c:d:816")) != -1 ) {
        switch (i) {
            case 'a':   if(optarg!=NULL) sscanf( optarg, "%x", &new_address );
                        break;
            case 'b':   if(optarg!=NULL) sscanf( optarg, "%x", &mod_address );
                        break;
            case 'c':   if(optarg!=NULL) {
                            sscanf( optarg, "%x", &rc_calibrate );
                            regime = RC_UPDATE;
                        }
                        break;
            case 'D':   if(optarg!=NULL) sscanf( optarg, "%s", device );
                        break;
            case 'd':   if(optarg!=NULL) {
                            sscanf( optarg, "%x", &high_dump );
                            regime = DUMP;
                        }
                        break;
            case 'N':   if(optarg!=NULL) sscanf( optarg, "%d", &new_serialno );
                        break;
            case 'R':   if(optarg!=NULL && sscanf( optarg, "%d", &new_rate ) ) new_ubrr = rate_to_ubrr(new_rate);
                        break;
            case 'r':   if(optarg!=NULL) sscanf( optarg, "%d", &ibaude_rate );
                        break;
            case 'f':   if(optarg!=NULL && sscanf( optarg, "%d", &fosc ) ) new_ubrr = rate_to_ubrr(new_rate);
                        break;
            case '8':   { prog_date = PROG_DATE_8; line_pars =	LINE_PARS_8; 
                          page_size = PAGE_SIZE_8; }
                        fprintf( stdout, "Using ATmega8 parameters...\n");
                        break;
            case '6':
                        fprintf( stdout, "Using ATmega16 parameters...\n");
            case '1':   { prog_date = PROG_DATE_16; line_pars =	LINE_PARS_16; 
                          page_size = PAGE_SIZE_16; }
                        break;
            case '?':   fprintf( stdout, "Option -%c is not legal..\n", optopt );
                        exit(1);
            case ':':   fprintf( stdout, "Option -%c has missing argument..\n", optopt );
            default:    exit(1);
        }
    }
    if( (regime != RC_UPDATE) && ((fname = av[optind]) == NULL) ) {
        fprintf( stdout, "Usage: avrup -options file.hex\n");
        fprintf( stdout, "\t-a ahex      set address ahex for application\n");
        fprintf( stdout, "\t-b bhex      update module with address bhex\n");
        fprintf( stdout, "\t-c rchex     update RC calibration constant rchex\n");
        fprintf( stdout, "\t-d hahex     dump flash memory\n");
        fprintf( stdout, "\t-N serialno  set device serial number within group\n");
        fprintf( stdout, "\t-R Rate      set exchange Rate for application\n");
        fprintf( stdout, "\t-r rate      update module using rate \n");
        fprintf( stdout, "\t-f freq      update module with nonstandard clock frequency \n");
        fprintf( stdout, "\t-8           use ATMega8 constants (default)\n");
        fprintf( stdout, "\t-16          use ATMega16 constants \n");
        fprintf( stdout, "\t-D device    use device instead of default (/dev/mcua0) \n");
        exit(1);
    }
    utc.current();
    fprintf( stdout, "Session %s %s\n", utc.strDate(), utc.strTime() );

    if( (regime == WRITE) && (datfile = fopen( fname, "r")) == NULL ) {
        fprintf( stdout, "Microcode not found\n");
        exit(1);
    }

    if( (regime == DUMP) && (datfile = fopen( fname, "w")) == NULL ) {
        fprintf( stdout, "Can't create file\n");
        exit(1);
    }

    try{
        RS485::initExchange( device );
        if( ibaude_rate > 0 ) RS485::setBaudeRate( ibaude_rate );
        switch( RS485::askData( mod_address, GET_IDENT, data) ){
            case APPLICATION:   fprintf( stdout, "Device is in application mode\n");
                                RS485::sendSimpleCommand( mod_address, REPROGRAMMING );
                                break;
            case UPDATE:        fprintf( stdout, "Device is in boot mode\n");
                                break;
            default:            fprintf( stdout, "Device doesn't respond...\n");
                                fprintf( stdout, "Check module address!..\n");
                                exit(1);
        }

        if( regime == RC_UPDATE ) {
            for( i = 0; i < FULL_SIZE; i++ ) code[i] = 0xffff;        
            code[RC_CORR] = rc_calibrate;
            adr_last = adr_begin = RC_CORR;
        }

        if( regime == DUMP ) {
            adr_begin = 0;
            adr_last = high_dump-1;
            fprintf( stdout, "Flash reading...\n");
            for( page = adr_begin/page_size; page < adr_last/page_size+1; page++ ){
                for( part = 0; part < page_size/PART_SIZE; part++) {
                    if( RS485::sendWordCommand( mod_address, SET_POINT, part+256*page ) &&
                        RS485::askData( mod_address, READ_FLASH, rdata) && 
                        part == rdata[0] && page == rdata[1] ) {
                        memcpy( (char*)buff, rdata+2, PART_LENGTH);
                        for( i=0; i < PART_SIZE; i++ )
                            fprintf( datfile, "%06x:%04x\n", page*page_size + part*PART_SIZE+i, buff[i] );
                    } else {
                        fprintf( stdout, "\nPage %02x part %02x read error !!! \n", page, part );
                        exit(1);
                    }
                }
            }
            RS485::sendSimpleCommand( mod_address, RESET );
            usleep(10000);
            if( datfile > 0 ) fclose( datfile );             
            exit(0);             
        } 

        if( regime == WRITE ) {
            fprintf( stdout, "Flash loading...\n");
                                // Code preparation
            for( i = 0; i < FULL_SIZE; i++ ) code[i] = 0xffff;
            while( fscanf( datfile, "%x:%x", &adr, &content) != EOF) {
                code[adr] = content;
                if(adr_begin==0xffff) adr_begin = adr;
                else unused+=adr-adr_last-1;
                adr_last = adr;
            }
                                    // Identification and line modification
            if( adr_begin == 0 ) {
                if( new_serialno ) code[prog_date-1] = (code[prog_date-1]&0xff00)|new_serialno;
                code[prog_date] = ( (utc.year()-2000)*16 + utc.month())*32 + utc.day();
                if( new_address ) code[line_pars] = (code[line_pars]&0xff00)|new_address;
                if( new_rate ) code[line_pars] = (code[line_pars]&0x00ff)|(new_ubrr<<8);
            }
        }
                                // Code upload
        for( page = adr_begin/page_size; page < adr_last/page_size+1; page++ ){
            acc=0xffff;
            for( i = 0; i < page_size; i++ ) {
                buff[i] = code[page_size*page+i];
                acc &= buff[i];
            }
            if( acc == 0xffff ) {
                printf("Page %02x  is empty\n", page );
                continue;
            }
            for( part = 0; part < page_size/PART_SIZE; part++) {
                i=PART_SIZE*part;
                data[0] = (unsigned char)part;
                data[1] = (unsigned char)page;
                memcpy( data+2, (char*)(buff+i), PART_LENGTH );
                if( RS485::sendData( mod_address, PART_LENGTH+2, data) &&
                    RS485::askData( mod_address, GET_PART, rdata) == PART_LENGTH+2 &&
                    memcmp( data, rdata, PART_LENGTH+2 ) == 0 ) {
                    fprintf( stdout, "Page %02x part %02x loading\r", page, part );
                    fflush( stdout );
                } else {
                    fprintf( stdout, "\nPage %02x part %02x error !!! \n", page, part );
                    exit(1);
                }
//  //            printf("%02x %02x %04x %04x %04x %04x %04x %04x %04x %04x\n", page, part, buff[i+0], buff[i+1], buff[i+2], buff[i+3], buff[i+4], buff[i+5], buff[i+6], buff[i+7] );
            }
            RS485::sendSimpleCommand( mod_address, WRITE_PAGE );
            usleep(10000);
            char status;
            if( !RS485::sendSimpleCommand( mod_address, TEST_PAGE ) ){
                RS485::askByte( mod_address, GET_STATUS, &status );
                fprintf( stdout, "\nPage %02x write error %02x !!! \n", page, status );
                exit(1);
            }
        }
                                // Code check
        fprintf( stdout, "\n%d words, %d unused\n", adr_last-adr_begin+1, unused );
        page = line_pars/page_size;
        part = (line_pars%page_size)/PART_SIZE;
        if( RS485::sendWordCommand( mod_address, SET_POINT, part+256*page ) &&
            RS485::askData( mod_address, READ_FLASH, rdata) && part == rdata[0] && page == rdata[1] ) {
                i = 2*(line_pars-page*page_size-part*PART_SIZE)+2;
                new_address = rdata[i];
		char subrr = rdata[i+1]; 
                new_ubrr = subrr;
//printf("!!! = %d  %d\n", new_ubrr, subrr);
        }
        RS485::sendSimpleCommand( mod_address, RESET );
        mod_address = new_address;
        usleep(10000);
        if( new_rate > 0 ) RS485::setBaudeRate( new_rate );
        if( RS485::askData( mod_address, GET_IDENT, data) == APPLICATION ) {
            fprintf( stdout, "Update done...\n");
            fprintf( stdout, "Address of the module: %02x\n", new_address);
            fprintf( stdout, "Exchange rate : %d Kbod\n", ubrr_to_rate(new_ubrr) );
            fprintf( stdout, "Now identification is: %02x %02x %02x %02x\n", data[0], data[1], data[2], data[3]);
        } else {
            fprintf( stdout, "Update failed or program code error...\n");
        }
    }
    catch( Error::Exchange err ){
        fprintf( stdout, "%s: %s\n", err.from, err.what );
        exit(1);
    }
    if( datfile > 0 ) fclose( datfile );
    exit(0);
}

